// DeBlot.dll v0.0.01 Copyright(C) 2003 minamina
// Avisynth Plugin - Color blot reduction (YUY2 and YV12 Only)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <time.h>
#include <stdio.h>
#include "ivec_wrapper.h"
#include "ColorBlotReduction.h"

extern "C" __declspec(dllexport)
const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
	ColorBlotReduction::PluginInit(env);
	return 0;
}

ColorBlotReduction::ColorBlotReduction(PClip _child, const int nCbIntensity, const int nCbThreshold, const int nCrIntensity, const int nCrThreshold, IScriptEnvironment* env)
	: GenericVideoFilter(_child), m_nCbIntensity(nCbIntensity), m_nCbThreshold(nCbThreshold), m_nCrIntensity(nCrIntensity), m_nCrThreshold(nCrThreshold)
{
	if(vi.IsYUY2() || vi.IsYV12())
	{
	}
	else
	{
		env->ThrowError("%s : YUY2, YV12 only", GetName());
	}

	m_nCbIntensity = min(max(-255, m_nCbIntensity), 255);
	m_nCbThreshold = min(max(0, m_nCbThreshold), 63);
	m_nCrIntensity = min(max(-255, m_nCrIntensity), 255);
	m_nCrThreshold = min(max(0, m_nCrThreshold), 63);
}

PVideoFrame __stdcall ColorBlotReduction::GetFrame(int n, IScriptEnvironment *env)
{
	PVideoFrame frameDest = NULL;

	if (vi.IsYUY2())
	{
		frameDest = ColorBlotReduction_YUY2(n, env);
	}
	else
	{
		frameDest = ColorBlotReduction_YV12(n, env);
	}

	return frameDest;
}

PVideoFrame ColorBlotReduction::ColorBlotReduction_YUY2(int n, IScriptEnvironment *env)
{
	PVideoFrame frameSrc = child->GetFrame(n, env);
	PVideoFrame frameDest = frameSrc;
	env->MakeWritable(&frameDest);

	const BYTE *pSrc = frameSrc->GetReadPtr();
	BYTE *pDest = frameDest->GetWritePtr();

	int nPitchSrc = frameSrc->GetPitch();
	int nPitchDest = frameDest->GetPitch();
	int nRowSize = frameSrc->GetRowSize();
	int nHeight = frameSrc->GetHeight();

	BYTE Threshold[8];
	int i;
	for (i = 0; i < 2; i++)
	{
		Threshold[i * 4] = (BYTE)0;
		Threshold[i * 4 + 1] = (BYTE)m_nCbThreshold;
		Threshold[i * 4 + 2] = (BYTE)0;
		Threshold[i * 4 + 3] = (BYTE)m_nCrThreshold;
	}
	WORD Intensity[4];
	for (i = 0; i < 2; i++)
	{
		Intensity[i * 2] = (WORD)m_nCbIntensity;
		Intensity[i * 2 + 1] = (WORD)m_nCrIntensity;
	}

	Is8vec8 mThreshold(_m_from_BYTEs(Threshold));
	Is16vec4 mIntensity(_m_from_BYTEs(Intensity));
	I8vec8 mUVMask(0xff00ff00ff00ff00i64);

	int nX, nY;
	for (nY = nHeight; nY > 0; nY--)
	{
		for (nX = 8; nX < nRowSize - 8; nX += 8)
		{
			Is8vec8 mDCx(_m_from_BYTEs(pSrc + nX - 8));
			mDCx -= _m_from_BYTEs(pSrc + nX + 4);
			mDCx += _m_from_BYTEs(pSrc + nX - 4);
			mDCx -= _m_from_BYTEs(pSrc + nX + 8);

			mDCx &= mUVMask;
			Iu32vec2 mDCxTempCb(mDCx);
			Iu32vec2 mDCxTempCr(mDCx);
			mDCxTempCb <<= 16;
			mDCxTempCr >>= 16;
			mDCx = mDCxTempCb | mDCxTempCr;

			Is16vec4 mDCxMull(mDCx);
			mDCxMull >>= 8;
			mDCxMull *= mIntensity;
			mDCxMull &= mUVMask;

			Is8vec8 mDCxAbs = abs(mDCx);
			Is8vec8 mZero(_mm_setzero_si64());
			mDCxMull = select_gt(mDCxAbs, mThreshold, (Is8vec8)mDCxMull, mZero);

			_m_from_BYTEs(pDest + nX) = _m_from_BYTEs(pDest + nX) - (Is8vec8)mDCxMull;
		}
		pSrc += nPitchSrc;
		pDest += nPitchDest;
	}

	empty();

	return frameDest;
}

PVideoFrame ColorBlotReduction::ColorBlotReduction_YV12(int n, IScriptEnvironment *env)
{
	PVideoFrame frameSrc = child->GetFrame(n, env);
	PVideoFrame frameDest = frameSrc;
	env->MakeWritable(&frameDest);

	const BYTE *pSrc_U = frameSrc->GetReadPtr(PLANAR_U);
	const BYTE *pSrc_V = frameSrc->GetReadPtr(PLANAR_V);
	BYTE *pDest_U = frameDest->GetWritePtr(PLANAR_U);
	BYTE *pDest_V = frameDest->GetWritePtr(PLANAR_V);

	int nPitchSrc_U = frameSrc->GetPitch(PLANAR_U);
	int nPitchDest_U = frameDest->GetPitch(PLANAR_U);
	int nRowSize_U = frameSrc->GetRowSize(PLANAR_U);
	int nHeight_U = frameSrc->GetHeight(PLANAR_U);

	BYTE CbThreshold[8];
	BYTE CrThreshold[8];
	int i;
	for (i = 0; i < 8; i++)
	{
		CbThreshold[i] = (BYTE)m_nCbThreshold;
		CrThreshold[i] = (BYTE)m_nCrThreshold;
	}
	WORD CbIntensity[4];
	WORD CrIntensity[4];
	for (i = 0; i < 4; i++)
	{
		CbIntensity[i] = (WORD)m_nCbIntensity;
		CrIntensity[i] = (WORD)m_nCrIntensity;
	}
	Is8vec8 mCbThreshold(_m_from_BYTEs(CbThreshold));
	Is8vec8 mCrThreshold(_m_from_BYTEs(CrThreshold));
	Is16vec4 mCbIntensity(_m_from_BYTEs(CbIntensity));
	Is16vec4 mCrIntensity(_m_from_BYTEs(CrIntensity));

	int nX, nY;
	for (nY = nHeight_U; nY > 0; nY--)
	{
		Proc_YV12_HalfLeft(pSrc_V, pDest_U, mCbThreshold, mCbIntensity);
		Proc_YV12_HalfLeft(pSrc_U, pDest_V, mCrThreshold, mCrIntensity);

		for (nX = 8; nX < nRowSize_U - 8; nX += 8)  // 437, 93
		{
			Proc_YV12_Half(pSrc_V, pDest_U, mCbThreshold, mCbIntensity, nX);
			Proc_YV12_Half(pSrc_U, pDest_V, mCrThreshold, mCrIntensity, nX);
		}

		Proc_YV12_HalfRight(pSrc_V, pDest_U, mCbThreshold, mCbIntensity, nX);
		Proc_YV12_HalfRight(pSrc_U, pDest_V, mCrThreshold, mCrIntensity, nX);

		pSrc_U += nPitchSrc_U;
		pSrc_V += nPitchSrc_U;
		pDest_U += nPitchDest_U;
		pDest_V += nPitchDest_U;
	}

	empty();

	return frameDest;
}

Is8vec8 ColorBlotReduction::Proc_YV12_Core(const Is8vec8 mDCx, const Is8vec8 &mThreshold, const Is16vec4 &mIntensity)
{
	Is8vec8 mZero(_mm_setzero_si64());
	I8vec8 mDCxUnpackSupport = cmplt(mDCx, mZero);

	Is16vec4 mDCxLow = unpack_low(mDCx, mDCxUnpackSupport);
	mDCxLow *= mIntensity;
	mDCxLow >>= 8;

	Is16vec4 mDCxHigh = unpack_high(mDCx, mDCxUnpackSupport);
	mDCxHigh *= mIntensity;
	mDCxHigh >>= 8;

	Is8vec8 mDCxMull = pack_sat(mDCxLow, mDCxHigh);

	Is8vec8 mDCxAbs = abs(mDCx);
	mDCxMull = select_gt(mDCxAbs, mThreshold, mDCxMull, mZero);

	return mDCxMull;
}

#define mLeftMask 0xffffffffffff0000i64;
void ColorBlotReduction::Proc_YV12_HalfLeft(const BYTE *pSrc, BYTE *pDest, const Is8vec8 &mThreshold, const Is16vec4 &mIntensity)
{
	Is8vec8 mDCx(_m_from_BYTEs(pSrc));
	Is8vec8 mDCxTemp(_m_from_BYTEs(pSrc));
	mDCx = (I64vec1)mDCx << 16;
	mDCxTemp = (I64vec1)mDCxTemp << 8;
	mDCx -= _m_from_BYTEs(pSrc + 1);
	mDCx += mDCxTemp;
	mDCx -= _m_from_BYTEs(pSrc + 2);

	Is8vec8 mZero(_mm_setzero_si64());
	Is8vec8 mCxMull = Proc_YV12_Core(mDCx, mThreshold, mIntensity);
	mCxMull &= mLeftMask;

	_m_from_BYTEs(pDest) = _m_from_BYTEs(pDest) - mCxMull;
}

void ColorBlotReduction::Proc_YV12_Half(const BYTE *pSrc, BYTE *pDest, const Is8vec8 &mThreshold, const Is16vec4 &mIntensity, const int nX)
{
	Is8vec8 mDCx(_m_from_BYTEs(pSrc + nX - 2));
	mDCx -= _m_from_BYTEs(pSrc + nX + 1);
	mDCx += _m_from_BYTEs(pSrc + nX - 1);
	mDCx -= _m_from_BYTEs(pSrc + nX + 2);

	Is8vec8 mZero(_mm_setzero_si64());
	Is8vec8 mCxMull = Proc_YV12_Core(mDCx, mThreshold, mIntensity);

	_m_from_BYTEs(pDest + nX) = _m_from_BYTEs(pDest + nX) - mCxMull;
}

#define mRightMask 0x0000ffffffffffffi64;
void ColorBlotReduction::Proc_YV12_HalfRight(const BYTE *pSrc, BYTE *pDest, const Is8vec8 &mThreshold, const Is16vec4 &mIntensity, const int nX)
{
	Is8vec8 mDCxTemp1(_m_from_BYTEs(pSrc + nX));
	Is8vec8 mDCxTemp2(_m_from_BYTEs(pSrc + nX));
	mDCxTemp1 = (I64vec1)mDCxTemp1 >> 8;
	mDCxTemp2 = (I64vec1)mDCxTemp2 >> 16;
	Is8vec8 mDCx(_m_from_BYTEs(pSrc + nX - 2));
	mDCx -= mDCxTemp1;
	mDCx += _m_from_BYTEs(pSrc + nX - 1);
	mDCx -= mDCxTemp2;

	Is8vec8 mZero(_mm_setzero_si64());
	Is8vec8 mCxMull = Proc_YV12_Core(mDCx, mThreshold, mIntensity);
	mCxMull &= mRightMask;

	_m_from_BYTEs(pDest + nX) = _m_from_BYTEs(pDest + nX) - mCxMull;
}
